"""Copyright (c) UChicago Argonne, LLC. All rights reserved.
See LICENSE file.
"""

import numpy as np

class RectROI:
    """A rectangular region of interest that can be applied to 3D dataset.
    
    
    """

    bounds = None
    calculation = None
    output = None
    
    def __init__(self, dims: list=None) -> None:

        if dims is None:
            self.bounds = {
                "x": (None, None),
                "y": (None, None),
                "z": (None, None)
            }
        else:
            if len(dims) != 3:
                raise ValueError("Invalid dims provided.")
            self.bounds = dict((dim, (None, None)) for dim in dims)
        
        self.calculation = {
            "output_data": None,
            "dims": None
        }

        self.output = {
            "data": None,
            "coords": None
        }
    
    def set_bounds(self, bounds: dict) -> None:
        """Sets coordinate bounds for the RectROI."""
        
        if type(bounds) != dict:
            raise ValueError("Invalid bounds provided.")
        if len(list(bounds.keys())) != 3:
            raise ValueError("Invalid bounds provided.")
        for dim in list(bounds.keys()):
            dim_bounds = bounds[dim]
            if type(dim_bounds) is None:
                bounds[dim] == (None, None)
            if type(dim_bounds) != list and type(dim_bounds) != tuple:
                raise ValueError("Invalid bounds provided.")
            
            if len(dim_bounds) != 2:
                raise ValueError("Invalid bounds provided.")
            if None not in bounds[dim] and dim_bounds[1] < dim_bounds[0]:
                raise ValueError("Invalid bounds provided.")

        if set(list(bounds.keys())) == set(list(self.bounds.keys())):
            self.bounds = {dim: bounds[dim] for dim in list(self.bounds.keys())}
        else:
            self.bounds = {dim: bounds[dim] for dim in list(bounds.keys())}

    def set_calculation(self, output: str, dims: list) -> None:
        """Sets the output calculation (average, max) and the dimensions to calculate on."""

        if dims is not None:
            if not set(list(self.bounds.keys())).issuperset(set(dims)):
                raise ValueError("Invalid dimension list provided.")
        
        if output not in ["average", "max"]:
            raise ValueError("Invalid output type provided. Accepted values are 'average' and 'max'.")
        
        self.calculation = {
            "output": output,
            "dims": dims
        }
    
    def apply(self, data, coords) -> None:
        """Carries out a calculation (see the 'output_type' attribute) on a dataset."""

        output_dims = self.calculation["dims"]
        output_type = self.calculation["output"]
        
        if output_dims is None:
            output_dims = []
        if output_type is None:
            raise ValueError("No output type found. Please add a output type using 'set_calculation'.")

        coords = coords.copy()

        # Find bounding pixels for ROI
        roi_idx = []
        roi_coords = {}
        for dim in list(coords.keys()):
            bound_1, bound_2 = None, None
            dim_coords = coords[dim]
            dim_bounds = self.bounds[dim]

            if dim_bounds[0] is None or np.searchsorted(dim_coords, dim_bounds[0]) == 0:
                if dim_bounds[1] is None or np.searchsorted(dim_coords, dim_bounds[1]) == len(dim_coords):
                    roi_idx.append(np.s_[:])
                    roi_coords.update({dim: dim_coords[np.s_[:]]})
                else:
                    bound_2 = np.searchsorted(dim_coords, dim_bounds[1])
                    roi_idx.append(np.s_[:bound_2])
                    roi_coords.update({dim: dim_coords[np.s_[:bound_2]]})
            else:
                bound_1 = np.searchsorted(dim_coords, dim_bounds[0])
                if dim_bounds[1] is None or np.searchsorted(dim_coords, dim_bounds[1]) == len(dim_coords):
                    roi_idx.append(np.s_[bound_1:])
                    roi_coords.update({dim: dim_coords[np.s_[bound_1:]]})
                else:
                    bound_2 = np.searchsorted(dim_coords, dim_bounds[1])
                    roi_idx.append(np.s_[bound_1:bound_2])
                    roi_coords.update({dim: dim_coords[np.s_[bound_1:bound_2]]})
        roi_data = data[tuple(roi_idx)]

        # Run output calculation
        if output_type == "average":

            if len(output_dims) == 0:
                raise ValueError("Dimension to average on not provided.")
            
            elif len(output_dims) == 1:
                avg_dim_idx = list(coords.keys()).index(output_dims[0])
                self.output["data"] = np.mean(roi_data, axis=avg_dim_idx)

                del(roi_coords[output_dims[0]])
                self.output["coords"] = roi_coords

            elif len(output_dims) == 2:
                avg_dim_idxs = [list(coords.keys()).index(dim) for dim in output_dims]
                self.output["data"] = np.mean(roi_data, axis=tuple(avg_dim_idxs))

                del(roi_coords[output_dims[0]])
                del(roi_coords[output_dims[1]])
                self.output["coords"] = roi_coords

            elif len(output_dims) == 3:
                self.output["data"] = np.mean(roi_data, axis=(0, 1, 2))

            else:
                raise ValueError("Invalid dimension list.")
            
        if output_type == "max":

            if len(output_dims) == 0:
                raise ValueError("Dimension to average on not provided.")
            
            elif len(output_dims) == 1:
                avg_dim_idx = list(coords.keys()).index(output_dims[0])
                self.output["data"] = np.amax(roi_data, axis=avg_dim_idx)

                del(roi_coords[output_dims[0]])
                self.output["coords"] = roi_coords

            elif len(output_dims) == 2:
                avg_dim_idxs = [list(coords.keys()).index(dim) for dim in output_dims]
                self.output["data"] = np.amax(roi_data, axis=tuple(avg_dim_idxs))

                del(roi_coords[output_dims[0]])
                del(roi_coords[output_dims[1]])
                self.output["coords"] = roi_coords

            elif len(output_dims) == 3:
                self.output["data"] = np.amax(roi_data, axis=(0, 1, 2))

            else:
                raise ValueError("Invalid dimension list.")

    def apply_to_scan(self, scan, data_type) -> None:
        
        if data_type == "raw":
            data = scan.raw_data["data"]
            coords = scan.raw_data["coords"]
        elif data_type == "gridded":
            data = scan.gridded_data["data"]
            coords = scan.gridded_data["coords"]
        else:
            raise("Invalid data type provided.")
        
        self.apply(data, coords)
    
    def get_output(self) -> dict:
        """Returns the output from the most recent apply() run."""
        
        return self.output



"""Copyright (c) UChicago Argonne, LLC. All rights reserved.
See LICENSE file.
"""

import databroker
import numpy as np
from prettytable import PrettyTable
import pyqtgraph as pg
import xrayutilities as xu

from xrdimageutil import utils
from xrdimageutil.gui import image_data_widget


class Catalog:
    """Houses functionality to filter and return xrdimageutil.Scan objects."""
    
    local_name = None
    bluesky_catalog = None # Bluesky dictionary-like catalog
    scan_uid_dict = None # Dictionary of scans in catalog with UID as key

    def __init__(self, local_name) -> None:

        self.local_name = str(local_name)
        self.bluesky_catalog = databroker.catalog[self.local_name]

        # Currently only configured for beamline 6-ID-B
        utils._add_catalog_handler(catalog=self)

        # Creates a Scan object for every run in the catalog
        # Adds Scans to a dictionary
        self.scan_uid_dict = dict(
            [(uid, Scan(catalog=self, uid=uid)) for uid in list(self.bluesky_catalog)]
        )

    def search(self, sample=None, proposal_id=None, user=None) -> list:
        """Returns a list of scan UID's from given criteria.
        
        This function is a limited extension of databroker's catalog search. 
        """

        query = {}
        if sample is not None:
            query.update({"sample": sample})
        if proposal_id is not None:
            query.update({"proposal_id": proposal_id})
        if user is not None:
            query.update({"user": user})

        search_results = self.bluesky_catalog.search(query)

        return search_results

    def list_scans(self) -> None:
        """Displays basic information about scans within the catalog.
        
        Depending on the size of the Catalog, this function may take 
        an extended time to run.
        """

        headers = ["scan_id", "motors", "sample", "proposal_id", "user"]
        table = PrettyTable(headers)

        for uid in list(self.scan_uid_dict.keys()):
            scan = self.scan_uid_dict[uid]
            row = [scan.scan_id, scan.motors, scan.sample, scan.proposal_id, scan.user]
            table.add_row(row)

        table.sortby = "scan_id"
        print(table)

    def get_scan(self, id) -> object:
        """Returns xrdimageutil.Scan object from given identifier.
        
        UID's and numerical scan ID's are both viable parameters.
        As for scan ID's, which are not necessarily unique, this 
        function will return the most recent Scan with the ID.

        Negative integers denoting recent scans also are acceptable.
        """

        # UID
        if type(id) == str:
            if id in self.scan_uid_dict.keys():
                return self.scan_uid_dict[id]
            else:
                raise KeyError(f"Scan with UID '{id}' not found.")

        # Scan ID
        elif type(id) == int:
            try:
                uid = self.bluesky_catalog[id].primary.metadata["start"]["uid"]
                return self.scan_uid_dict[uid]
            except ValueError:
                raise KeyError(f"Scan with ID #{id} not found.")

        else:
            raise TypeError(f"Scan ID must be either str or int.")

    def get_scans(self, ids: list) -> list:
        """Returns xrdimageutil.Scan objects from list of given identifiers.
        
        See xrdimageutil.Catalog.get_scan for more details.
        """

        if type(ids) != list:
            raise TypeError("Input needs to be a list.")

        scan_list = []
        for id in ids:
            scan = self.get_scan(id=id)
            scan_list.append(scan)

        return scan_list

class Scan(object):
    """Houses data from a singular Bluesky run and additional image data."""

    catalog = None # Parent Catalog
    uid = None # UID for scan; given by bluesky

    bluesky_run = None # Raw Bluesky run for scan

    scan_id = None # Numerical ID -- not always unique
    sample = None # Experimental sample
    proposal_id = None # Manually provided Proposal ID
    user = None # Experimental user
    motors = None # List of variable motors for scan
    
    rsm = None # Reciprocal space map for every point within a scan
    raw_data = None # Data and coords for raw 2D detector images
    gridded_data = None # Data and coords for mapped 3D image

    def __init__(self, catalog: Catalog, uid: str) -> None:

        self.catalog = catalog
        self.uid = uid

    def __getattribute__(self, __name: str):
        """Lazy loading for class variables."""

        if __name == "bluesky_run":
            if object.__getattribute__(self, __name) is None:
                object.__setattr__(self, __name, self.catalog.bluesky_catalog[self.uid])
            return object.__getattribute__(self, __name)
        elif __name == "scan_id":
            if object.__getattribute__(self, __name) is None:
                object.__setattr__(self, __name, self.bluesky_run.metadata["start"]["scan_id"])
            return object.__getattribute__(self, __name) 
        elif __name == "sample":
            if object.__getattribute__(self, __name) is None:
                object.__setattr__(self, __name, self.bluesky_run.metadata["start"]["sample"])
            return object.__getattribute__(self, __name) 
        elif __name == "proposal_id":
            if object.__getattribute__(self, __name) is None:
                object.__setattr__(self, __name, self.bluesky_run.metadata["start"]["proposal_id"])
            return object.__getattribute__(self, __name)
        elif __name == "user":
            if object.__getattribute__(self, __name) is None:
                object.__setattr__(self, __name, self.bluesky_run.metadata["start"]["user"])
            return object.__getattribute__(self, __name)
        elif __name == "motors":
            if object.__getattribute__(self, __name) is None:
                object.__setattr__(self, __name, self.bluesky_run.metadata["start"]["motors"])
            return object.__getattribute__(self, __name)
        elif __name == "rsm":
            if object.__getattribute__(self, __name) is None:
                object.__setattr__(self, __name, utils._get_rsm_for_scan(self))
            return object.__getattribute__(self, __name)
            if object.__getattribute__(self, __name) is None:
                object.__setattr__(self, __name, utils._get_rsm_bounds(self))
            return object.__getattribute__(self, __name)
        elif __name == "raw_data":
            if object.__getattribute__(self, __name) is None:
                object.__setattr__(self, __name, utils._get_raw_data(self))
            return object.__getattribute__(self, __name)
        elif __name == "gridded_data":
            if object.__getattribute__(self, __name) is None:
                object.__setattr__(self, __name, {"data": None, "coords": None})
            return object.__getattribute__(self, __name)
        else:
            return object.__getattribute__(self, __name)
        
    def grid_data(self, shape: tuple, bounds: dict=None) -> None:
        """Creates a gridded 3D image according to coordinates given by a Scan's reciprocal space map."""

        # Shape validation
        if type(shape) is not tuple:
            raise TypeError(f"Shape must be a tuple.")
        if len(shape) != 3:
            raise ValueError(f"Shape must be of length 3.")
        for i in shape:
            if type(i) != int:
                raise ValueError(f"Dimension shape must consist of integers.")
            if i < 10:
                raise ValueError(f"Minimum gridded data shape is (10, 10, 10).")
        
        # Bounds validation
        if bounds is None:
            bounds = {
                "H": (np.amin(self.rsm[:,:,:,0]), np.amax(self.rsm[:,:,:,0])),
                "K": (np.amin(self.rsm[:,:,:,1]), np.amax(self.rsm[:,:,:,1])),
                "L": (np.amin(self.rsm[:,:,:,2]), np.amax(self.rsm[:,:,:,2]))
            }
        else:
            if set(list(bounds.keys())) != set(["H", "K", "L"]):
                raise ValueError(f"Expects 'H', 'K', and 'L' as keys for gridded data bounds.")
            for i in list(bounds.keys()):
                if type(bounds[i]) != tuple or type(bounds[i]) != list or len(bounds[i]) != 2:
                    raise ValueError(f"Expects a tuple/list (len 2) denoting a min and max value for each dimension.")
                if bounds[i][0] >= bounds[i][1]:
                    raise ValueError(f"First bound must be less than second bound.")
                
        # Prepares gridder bounds/interpolation
        gridder = xu.Gridder3D(
            nx=shape[0], 
            ny=shape[1], 
            nz=shape[2]
        )
        gridder.KeepData(True)
        gridder.dataRange(
            xmin=bounds["H"][0], xmax=bounds["H"][1],
            ymin=bounds["K"][0], ymax=bounds["K"][1],
            zmin=bounds["L"][0], zmax=bounds["L"][1],
            fixed=True
        )

        # Grids raw data with bounds
        gridder(
            self.rsm[:,:,:,0], 
            self.rsm[:,:,:,1], 
            self.rsm[:,:,:,2], 
            self.raw_data["data"]
        )
        self.gridded_data["data"] = gridder.data

        # Retrieves HKL coordinates for gridded data
        self.gridded_data["coords"] = {
            "H": gridder.xaxis, 
            "K": gridder.yaxis, 
            "L": gridder.zaxis
        }
    
    def point_count(self) -> int:
        """Returns the number of points in a Scan."""

        if "primary" not in self.bluesky_run.keys():
            return 0
        elif "dims" not in self.bluesky_run.primary.metadata.keys():
            return 0
        else:
            return self.bluesky_run.primary.metadata["dims"]["time"]

    def view_image_data(self) -> None:
        """Displays Scan image data in an interactive GUI."""
        
        self.app = pg.mkQApp()
        self.gui_window = image_data_widget.ScanImageDataGUI(scan=self)
        self.gui_window.raise_()
        self.gui_window.show()
        self.gui_window.raise_()
        self.app.exec_()
